/*
  Allwinner.h - GPIO and PWM functions

  Allwinner H2+, H3, H5, H6 and A10, A20, A23, A64, A83T Supported.

  Copyright (c) 2020 Sasapea's Lab. All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/
#ifndef __ALLWINNER_H
#define __ALLWINNER_H

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdint.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/mman.h>

// Select H Serise = 3 or 5 or 6
// Select A Serise = 10 or 20 or 23 or 64 or 83
#if !defined(ALLWINNER_H) && !defined(ALLWINNER_A)
#define ALLWINNER_H 3
#endif

/************************************************************/
/* GPIO                                                     */
/************************************************************/

#if (ALLWINNER_A == 83)

#define GPIO0_BASE_ADDRESS 0x01C20800
#define GPIO1_BASE_ADDRESS 0x01F02C00
#define GPIO0_EINT_REG {0, 1, 0, 0, 0, 0, 2, 3, 0, 0, 0} // B, G, H
#define GPIO1_EINT_REG {1, 0}                            // L

#elif (ALLWINNER_A == 64)

#define GPIO0_BASE_ADDRESS 0x01C20800
#define GPIO0_EINT_REG {0, 1, 0, 0, 0, 0, 2, 3, 0, 0, 0} // B, G, H

#elif (ALLWINNER_A == 23)

#define GPIO0_BASE_ADDRESS 0x01C20800
#define GPIO0_EINT_REG {1, 2, 0, 0, 0, 0, 3, 0, 0, 0, 0} // A, B, G

#elif (ALLWINNER_A == 20)

#define GPIO0_BASE_ADDRESS 0x01C20800
#define GPIO0_EINT_REG {1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0} // Port Mixed

#elif (ALLWINNER_A == 10)

#define GPIO0_BASE_ADDRESS 0x01C20800
#define GPIO0_EINT_REG {1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0} // Port Mixed

#elif (ALLWINNER_H == 6)

#define GPIO0_BASE_ADDRESS 0x0300B000
#define GPIO1_BASE_ADDRESS 0x07022000
#define GPIO0_EINT_REG {0, 0, 0, 0, 0, 1, 2, 3, 0, 0, 0} // F, G, H
#define GPIO1_EINT_REG {1, 2}                            // L, M

#elif (ALLWINNER_H == 5)

#define GPIO0_BASE_ADDRESS 0x01C20800
#define GPIO1_BASE_ADDRESS 0x01F02C00
#define GPIO0_EINT_REG {1, 0, 0, 0, 0, 2, 3, 0, 0, 0, 0} // A, F, G
#define GPIO1_EINT_REG {1, 0}                            // L

#else // H3 and H2+

#define GPIO0_BASE_ADDRESS 0x01C20800
#define GPIO1_BASE_ADDRESS 0x01F02C00
#define GPIO0_EINT_REG {1, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0} // A, G
#define GPIO1_EINT_REG {1, 0}                            // L

#endif

typedef enum
{
  GPIO_FUNC_INPUT,
  GPIO_FUNC_OUTPUT,
  GPIO_FUNC_2,
  GPIO_FUNC_3,
  GPIO_FUNC_4,
  GPIO_FUNC_5,
  GPIO_FUNC_EINT,
  GPIO_FUNC_DISABLE,
} GPIO_FUNC;

typedef enum
{
  GPIO_DRIVE_LEVEL0,
  GPIO_DRIVE_LEVEL1,
  GPIO_DRIVE_LEVEL2,
  GPIO_DRIVE_LEVEL3,
} GPIO_DRIVE;

typedef enum
{
  GPIO_PULL_DISABLE,
  GPIO_PULL_UP,
  GPIO_PULL_DOWN,
} GPIO_PULL;

typedef enum
{
  GPIO_INT_CONF_POSITIVE_EDGE,
  GPIO_INT_CONF_NEGATIVE_EDGE,
  GPIO_INT_CONF_HIGH_LEVEL,
  GPIO_INT_CONF_LOW_LEVEL,
  GPIO_INT_CONF_DOUBLE_EDGE,
} GPIO_INT_CONF;

typedef enum
{
  GPIO_INT_CTRL_DISABLE,
  GPIO_INT_CTRL_ENABLE,
} GPIO_INT_CTRL;

typedef enum
{
  GPIO_INT_STATUS_NO_IRQ_PENDING,
  GPIO_INT_STATUS_IRQ_PENDING
} GPIO_INT_STATUS;


typedef struct
{
  volatile uint32_t cfg[4];
  volatile uint32_t dat;
  volatile uint32_t drv[2];
  volatile uint32_t pul[2];
} GPIO_DCR;

typedef struct
{
  volatile uint32_t cfg[4];
  volatile uint32_t ctl;
  volatile uint32_t sta;
  volatile uint32_t deb;
  volatile uint32_t rsv;
} GPIO_ICR;

typedef struct
{
  GPIO_DCR dcr[0x0E]; // 0x0000 - 0x01F8 (0x01F8)
  uint32_t rsv[0x02]; // 0x01F8 - 0x0200 (0x0008)
  GPIO_ICR icr[0x10]; // 0x0200 - 0x0400 (0x0200)
} GPIO_REG;

class Allwinner_PIO
{
  private:

    GPIO_REG *_gpio0;
    GPIO_REG *_gpio1;

    GPIO_DCR *getDCR(uint16_t pin)
    {
      static const uint8_t PORT0[] = GPIO0_EINT_REG;
#ifdef GPIO1_EINT_REG
      static const uint8_t PORT1[] = GPIO1_EINT_REG;
#endif
      if ((pin >>= 5) < sizeof(PORT0) / sizeof(PORT0[0]))
      {
        if (_gpio0)
          return &_gpio0->dcr[pin];
      }
#ifdef GPIO1_EINT_REG
      else if ((pin -= sizeof(PORT0) / sizeof(PORT0[0])) < sizeof(PORT1) / sizeof(PORT1[0]))
      {
        if (_gpio1)
          return &_gpio1->dcr[pin];
      }
#endif
      return NULL;
    }

    GPIO_ICR *getICR(uint16_t pin)
    {
      static const uint8_t PORT0[] = GPIO0_EINT_REG;
#ifdef GPIO1_EINT_REG
      static const uint8_t PORT1[] = GPIO1_EINT_REG;
#endif
      if ((pin >>= 5) < sizeof(PORT0) / sizeof(PORT0[0]))
      {
        pin = PORT0[pin];
        if (_gpio0 && pin)
          return &_gpio0->icr[pin - 1];
      }
#ifdef GPIO1_EINT_REG
      else if ((pin -= sizeof(PORT0) / sizeof(PORT0[0])) < sizeof(PORT1) / sizeof(PORT1[0]))
      {
        pin = PORT1[pin];
        if (_gpio1 && pin)
          return &_gpio1->icr[pin - 1];
      }
#endif
      return NULL;
    }

  public:

    Allwinner_PIO(void)
    : _gpio0(NULL)
    , _gpio1(NULL)
    {
      int fd = open("/dev/mem", O_RDWR);
      if (fd != -1)
      {
        unsigned int ps = sysconf(_SC_PAGESIZE);
        void *mm = mmap(0, ps, PROT_READ | PROT_WRITE, MAP_SHARED, fd, GPIO0_BASE_ADDRESS & ~(ps - 1));
        if (mm == MAP_FAILED)
          fprintf(stderr, "[ERROR] mmap(0, %d, PROT_READ | PROT_WRITE, MAP_SHARED, %d, %08X) = %s\n",
            ps, fd, GPIO0_BASE_ADDRESS & ~(ps - 1), strerror(errno));
        else
          _gpio0 = (GPIO_REG *)((uint8_t *)mm + (GPIO0_BASE_ADDRESS & (ps - 1)));
#ifdef GPIO1_BASE_ADDRESS
        mm = mmap(0, ps, PROT_READ | PROT_WRITE, MAP_SHARED, fd, GPIO1_BASE_ADDRESS & ~(ps - 1));
        if (mm == MAP_FAILED)
          fprintf(stderr, "[ERROR] mmap(0, %d, PROT_READ | PROT_WRITE, MAP_SHARED, %d, %08X) = %s\n",
            ps, fd, GPIO1_BASE_ADDRESS & ~(ps - 1), strerror(errno));
        else
          _gpio1 = (GPIO_REG *)((uint8_t *)mm + (GPIO1_BASE_ADDRESS & (ps - 1)));
#endif
        close(fd);
      }
    }

    virtual ~Allwinner_PIO(void)
    {
    }

    void begin(void)
    {
    }

    void end(void)
    {
    }

    GPIO_FUNC func(uint16_t pin, GPIO_FUNC mode)
    {
      GPIO_FUNC rv = GPIO_FUNC_INPUT;
      GPIO_DCR *dcr = getDCR(pin);
      if (dcr)
      {
        volatile uint32_t *reg = &dcr->cfg[(pin >> 3) & 0x03];
        int bit = (pin & 0x07) << 2;
        rv = (GPIO_FUNC)((*reg >> bit) & 0x0F);
        *reg = (*reg & ~(0x0F << bit)) | ((mode & 0x0F) << bit);
      }
      return rv;
    }

    GPIO_PULL pull(uint16_t pin, GPIO_PULL mode)
    {
      GPIO_PULL rv = GPIO_PULL_DISABLE;
      GPIO_DCR *dcr = getDCR(pin);
      if (dcr)
      {
        volatile uint32_t *reg = &dcr->pul[(pin >> 4) & 0x01];
        int bit = (pin & 0x0F) << 1;
        rv = (GPIO_PULL)((*reg >> bit) & 0x03);
        *reg = (*reg & ~(0x03 << bit)) | ((mode & 0x03) << bit);
      }
      return rv;
    }

    GPIO_DRIVE drive(uint16_t pin, GPIO_DRIVE mode)
    {
      GPIO_DRIVE rv = GPIO_DRIVE_LEVEL1;
      GPIO_DCR *dcr = getDCR(pin);
      if (dcr)
      {
        volatile uint32_t *reg = &dcr->drv[(pin >> 4) & 0x01];
        int bit = (pin & 0x0F) << 1;
        rv = (GPIO_DRIVE)((*reg >> bit) & 0x03);
        *reg = (*reg & ~(0x03 << bit)) | ((mode & 0x03) << bit);
      }
      return rv;
    }

    void write(uint16_t pin, uint8_t value)
    {
      GPIO_DCR *dcr = getDCR(pin);
      if (dcr)
      {
        int bit = 1 << (pin & 0x1F);
        if (value)
          dcr->dat |=  bit;
        else
          dcr->dat &= ~bit;
      }
    }

    uint8_t read(uint16_t pin)
    {
      GPIO_DCR *dcr = getDCR(pin);
      return dcr ? (dcr->dat >> (pin & 0x1F)) & 1 : 0;
    }

    void interrput_config(uint16_t pin, GPIO_INT_CONF mode)
    {
      GPIO_ICR *icr = getICR(pin);
      if (icr)
      {
        volatile uint32_t *reg = &icr->cfg[(pin & 0x1F) >> 3];
        int bit = (pin & 0x07) << 2;
        *reg = (*reg & ~(0x0F << bit)) | ((mode & 0x0F) << bit);
      }
    }

    void interrput_enable(uint16_t pin, GPIO_INT_CTRL value)
    {
      GPIO_ICR *icr = getICR(pin);
      if (icr)
      {
        int bit = 1 << (pin & 0x1F);
        if (value)
          icr->ctl |=  bit;
        else
          icr->ctl &= ~bit;
      }
    }

    GPIO_INT_STATUS interrput_status(uint16_t pin)
    {
      GPIO_ICR *icr = getICR(pin);
      if (icr)
      {
        int bit = 1 << (pin & 0x1F);
        if (icr->sta & bit)
        {
          icr->sta |= bit;
          return GPIO_INT_STATUS_IRQ_PENDING;
        }
      }
      return GPIO_INT_STATUS_NO_IRQ_PENDING;
    }

    void interrput_debounce(uint16_t pin, uint8_t deb_clk_pre_scale, uint8_t pio_int_clk_select)
    {
      GPIO_ICR *icr = getICR(pin);
      if (icr)
        icr->deb = ((deb_clk_pre_scale & 0x03) << 4) | (pio_int_clk_select & 1);
    }
};

extern Allwinner_PIO GPIO;

/************************************************************/
/* PWM                                                      */
/************************************************************/

#include "PWMBase.h"

#if (ALLWINNER_A == 83)

#define PWM0_BASE_ADDRESS 0x01C21400
#define PWM0_PORT         ((('D' - 'A') << 5) + 28) // PD28
#define PWM0_SELECT       GPIO_FUNC_2
#define PWM1_BASE_ADDRESS 0
#define PWM1_PORT         PWM0_PORT
#define PWM1_SELECT       PWM0_SELECT

#elif (ALLWINNER_A == 64)

#define PWM0_BASE_ADDRESS 0x01C21400
#define PWM0_PORT         ((('D' - 'A') << 5) + 22) // PD22
#define PWM0_SELECT       GPIO_FUNC_2
#define PWM1_BASE_ADDRESS 0
#define PWM1_PORT         PWM0_PORT
#define PWM1_SELECT       PWM0_SELECT

#elif (ALLWINNER_A == 23)

#define PWM0_BASE_ADDRESS 0x01C21400
#define PWM0_PORT         ((('H' - 'A') << 5) + 0) // PH00
#define PWM0_SELECT       GPIO_FUNC_2
#define PWM1_BASE_ADDRESS 0x01C21400
#define PWM1_PORT         ((('H' - 'A') << 5) + 1) // PH01
#define PWM1_SELECT       GPIO_FUNC_2

#elif (ALLWINNER_A == 20)

#define PWM0_BASE_ADDRESS 0x01C20C00
#define PWM0_PORT         ((('B' - 'A') << 5) + 2) // PB02
#define PWM0_SELECT       GPIO_FUNC_2
#define PWM1_BASE_ADDRESS 0x01C20C00
#define PWM1_PORT         ((('I' - 'A') << 5) + 3) // PI03
#define PWM1_SELECT       GPIO_FUNC_2

#elif (ALLWINNER_A == 10)

#define PWM0_BASE_ADDRESS 0x01C20C00
#define PWM0_PORT         ((('B' - 'A') << 5) + 2) // PB02
#define PWM0_SELECT       GPIO_FUNC_2
#define PWM1_BASE_ADDRESS 0x01C20C00
#define PWM1_PORT         ((('I' - 'A') << 5) + 3) // PI03
#define PWM1_SELECT       GPIO_FUNC_2

#elif (ALLWINNER_H == 6)

#define PWM0_BASE_ADDRESS 0x0300A000
#define PWM0_PORT         ((('D' - 'A') << 5) + 22) // PD22
#define PWM0_SELECT       GPIO_FUNC_2
#define PWM1_BASE_ADDRESS 0
#define PWM1_PORT         PWM0_PORT
#define PWM1_SELECT       PWM0_SELECT

#else // H5, H3 and H2+

#define PWM0_BASE_ADDRESS 0x01C21400
#define PWM0_PORT         ((('A' - 'A') << 5) +  5) // PA05
#define PWM0_SELECT       GPIO_FUNC_3
#define PWM1_BASE_ADDRESS 0
#define PWM1_PORT         PWM0_PORT
#define PWM1_SELECT       PWM0_SELECT

#endif

#define PWM_SCLK 24000000

#define PWM1_RDY_POS           29
#define PWM0_RDY_POS           28
#define PWM1_BYPASS_POS        24
#define PWM1_PUL_START_POS     23
#define PWM1_MODE_POS          22
#define PWM1_SCLK_GATING_POS   21
#define PWM1_ACT_STA_POS       20
#define PWM1_EN_POS            19
#define PWM1_PRESCAL_POS       15
#define PWM0_BYPASS_POS        9
#define PWM0_PUL_START_POS     8
#define PWM0_MODE_POS          7
#define PWM0_SCLK_GATING_POS   6
#define PWM0_ACT_STA_POS       5
#define PWM0_EN_POS            4
#define PWM0_PRESCAL_POS       0

#define PWM_PRESCAL_LEN        4
#define PWM_PRESCAL_MSK        ((1 << PWM_PRESCAL_LEN) - 1)
#define PWM_ENTIRE_CYS_POS     16
#define PWM_ENTIRE_CYS_LEN     16
#define PWM_ENTIRE_CYS_MSK     ((1 << PWM_ENTIRE_CYS_LEN) - 1)
#define PWM_ENTIRE_ACT_CYS_POS 0
#define PWM_ENTIRE_ACT_CYS_LEN 16
#define PWM_ENTIRE_ACT_CYS_MSK ((1 << PWM_ENTIRE_ACT_CYS_LEN) - 1)

#define PWM_MODE_CYCLE 0
#define PWM_MODE_PULS  1

#define PWM_MIN_FRQUENCY 1
#define PWM_MAX_FRQUENCY PWM_SCLK

typedef struct
{
  uint32_t  RDY;
  uint32_t  BYPASS;
  uint32_t  PUL_START;
  uint32_t  MODE;
  uint32_t  SCLK_GATING;
  uint32_t  ACT_STA;
  uint32_t  EN;
  uint32_t  PRESCAL;
  GPIO_FUNC SELECT;
} PWM_CHANNEL;

typedef struct
{
  volatile uint32_t ctrl;
  volatile uint32_t period[2];
} PWM_REG;

class Allwinner_PWM : public PWMBase
{
  private:

    PWM_REG    *_pwm;
    PWM_CHANNEL _ch;
    uint32_t    _num;
    uint32_t    _ctrl;
    uint32_t    _count;
    uint32_t    _pscal;
    GPIO_FUNC   _func;

    void prescal(void)
    {
      static const uint32_t CLKS[] = {PWM_SCLK / 120, PWM_SCLK / 180, PWM_SCLK / 240, PWM_SCLK / 360, PWM_SCLK / 480,
        0, 0, 0, PWM_SCLK / 12000, PWM_SCLK / 24000, PWM_SCLK / 36000, PWM_SCLK / 48000, PWM_SCLK / 72000};
      if (_frequency)
      {
        _pscal = -1;
        _count = PWM_SCLK / _frequency;
        while (_count > PWM_ENTIRE_CYS_MSK)
        {
          if (_pscal + 1 >= sizeof(CLKS) / sizeof(CLKS[0]))
          {
            _count = PWM_ENTIRE_CYS_MSK;
            break;
          }
          if (CLKS[++_pscal])
            _count = CLKS[_pscal] / _frequency;
        }
        if (_count == 0)
          _count = 1;
        _pscal &= PWM_PRESCAL_MSK;
      }
      else
      {
        _pscal = sizeof(CLKS) / sizeof(CLKS[0]) - 1;
        _count = PWM_ENTIRE_CYS_MSK;
      }
    }

    void setup(void)
    {
      if (_pwm)
      {
        // wait busy
        while ((_pwm->ctrl & _ch.RDY) == _ch.RDY)
          continue;
        uint32_t ps = (_pscal << _ch.PRESCAL);
        // puls mode
        if (_ctrl & _ch.MODE)
        {
          _pwm->ctrl = _ctrl | ps | _ch.PUL_START;
          _pwm->period[_num] = _count << PWM_ENTIRE_ACT_CYS_POS;
        }
        // cycle mode
        else
        {
          if ((_count == 1) && (_duty >= PWM_DUTY_50) && (_duty < PWM_DUTY_100))
            _pwm->ctrl = _ctrl | ps | _ch.BYPASS;
          else
            _pwm->ctrl = _ctrl | ps;
          _pwm->period[_num] = ((_count - 1) << PWM_ENTIRE_CYS_POS) | ((_count * _duty / PWM_DUTY_100) << PWM_ENTIRE_ACT_CYS_POS);
        }
      }
    }

    void enable(uint8_t value)
    {
      uint32_t flag = _ch.SCLK_GATING | _ch.EN;
      if (value)
        _ctrl |=  flag;
      else
        _ctrl &= ~flag;
      setup();
    }

    void mode(uint8_t value)
    {
      if (value)
        _ctrl |=  _ch.MODE;
      else
        _ctrl &= ~_ch.MODE;
    }

  public:

    Allwinner_PWM(uint8_t num)
    : PWMBase(num ? PWM1_PORT : PWM0_PORT)
    , _pwm(NULL)
    , _ctrl(0)
    , _func(GPIO_FUNC_INPUT)
    {
      static const uint32_t BASE[] = {PWM0_BASE_ADDRESS, PWM1_BASE_ADDRESS};
      uint32_t base = BASE[num];
      if (base)
      {
        int fd = open("/dev/mem", O_RDWR);
        if (fd != -1)
        {
          unsigned int ps = sysconf(_SC_PAGESIZE);
          void *mm = mmap(0, ps, PROT_READ | PROT_WRITE, MAP_SHARED, fd, PWM0_BASE_ADDRESS & ~(ps - 1));
          if (mm == MAP_FAILED)
            fprintf(stderr, "[ERROR] mmap(0, %d, PROT_READ | PROT_WRITE, MAP_SHARED, %d, %08X) = %s\n",
              ps, fd, PWM0_BASE_ADDRESS & ~(ps - 1), strerror(errno));
          else
            _pwm = (PWM_REG *)((uint8_t *)mm + (PWM0_BASE_ADDRESS & (ps - 1)));
          close(fd);
        }
      }
      if (_num)
      {
        _ch.RDY         = (1 << PWM1_RDY_POS) | (1 << PWM1_EN_POS);
        _ch.BYPASS      = (1 << PWM1_BYPASS_POS);
        _ch.PUL_START   = (1 << PWM1_PUL_START_POS);
        _ch.MODE        = (1 << PWM1_MODE_POS);
        _ch.SCLK_GATING = (1 << PWM1_SCLK_GATING_POS);
        _ch.ACT_STA     = (1 << PWM1_ACT_STA_POS);
        _ch.EN          = (1 << PWM1_EN_POS);
        _ch.PRESCAL     = PWM1_PRESCAL_POS;
        _ch.SELECT      = PWM1_SELECT;
      }
      else
      {
        _ch.RDY         = (1 << PWM0_RDY_POS) | (1 << PWM0_EN_POS);
        _ch.BYPASS      = (1 << PWM0_BYPASS_POS);
        _ch.PUL_START   = (1 << PWM0_PUL_START_POS);
        _ch.MODE        = (1 << PWM0_MODE_POS);
        _ch.SCLK_GATING = (1 << PWM0_SCLK_GATING_POS);
        _ch.ACT_STA     = (1 << PWM0_ACT_STA_POS);
        _ch.EN          = (1 << PWM0_EN_POS);
        _ch.PRESCAL     = PWM0_PRESCAL_POS;
        _ch.SELECT      = PWM0_SELECT;
      }
      prescal();
    }

    virtual ~Allwinner_PWM(void)
    {
    }

    void begin(int value = 0)
    {
      if ((_ctrl & _ch.EN) == 0)
      {
        mode(value);
        enable(true);
        if (_pwm)
          _func = GPIO.func(_pin, _ch.SELECT);
      }
    }

    void end(void)
    {
      if (_ctrl & _ch.EN)
      {
        if (_pwm)
          GPIO.func(_pin, _func);
        enable(false);
      }
    }

    void polarity(int value)
    {
      uint32_t save = _ctrl;
      if (value)
        _ctrl |=  _ch.ACT_STA; // active high
      else
        _ctrl &= ~_ch.ACT_STA; // active low
      if ((save ^ _ctrl) & _ch.ACT_STA)
        setup();
    }

    void pulse(uint32_t ns)
    {
      if (ns == 0)
        ns = 1;
      ns = 1000000000 / ns;
      if (_frequency != ns)
      {
        _frequency = ns;
        prescal();
        setup();
      }
    }

    void frequency(uint32_t value)
    {
      if (value < PWM_MIN_FRQUENCY)
        value = PWM_MIN_FRQUENCY;
      else if (value > PWM_MAX_FRQUENCY)
        value = PWM_MAX_FRQUENCY;
      if (_frequency != value)
      {
        PWMBase::frequency(value);
        prescal();
        setup();
      }
    }

    void duty(uint32_t value)
    {
      if (value > PWM_DUTY_100)
        value = PWM_DUTY_100;
      if (_duty != value)
      {
        PWMBase::duty(value);
        setup();
      }
    }
};

extern Allwinner_PWM PWM0[2];

/************************************************************/
/* TWI                                                      */
/************************************************************/

#if (ALLWINNER_A == 83)

#define TWI0_BASE_ADDRESS 0x01C2AC00 // PH00-PH01
#define TWI1_BASE_ADDRESS 0x01C2B000 // PH02-PH03
#define TWI2_BASE_ADDRESS 0x01C2B400 // PH04-PH05
#define TWI3_BASE_ADDRESS 0x01F02400 // PL00-PL01
#define TWI4_BASE_ADDRESS 0

#elif (ALLWINNER_A == 64)

#define TWI0_BASE_ADDRESS 0x01C2AC00 // PH00-PH01
#define TWI1_BASE_ADDRESS 0x01C2B000 // PH02-PH03
#define TWI2_BASE_ADDRESS 0x01C2B400 // PE14-PE15
#define TWI3_BASE_ADDRESS 0
#define TWI4_BASE_ADDRESS 0

#elif (ALLWINNER_A == 23)

#define TWI0_BASE_ADDRESS 0x01C2AC00 // PH02-PH03
#define TWI1_BASE_ADDRESS 0x01C2B000 // PH04-PH05
#define TWI2_BASE_ADDRESS 0x01C2B400 // PE12-PE13
#define TWI3_BASE_ADDRESS 0
#define TWI4_BASE_ADDRESS 0

#elif (ALLWINNER_A == 20)

#define TWI0_BASE_ADDRESS 0x01C2AC00 // PB00-PB01
#define TWI1_BASE_ADDRESS 0x01C2B000 // PB18-PB19
#define TWI2_BASE_ADDRESS 0x01C2B400 // PB20-PB21
#define TWI3_BASE_ADDRESS 0x01C2B800 // PL00-PL01
#define TWI4_BASE_ADDRESS 0x01C2C000 // PL02-PL03

#elif (ALLWINNER_A == 10)

#define TWI0_BASE_ADDRESS 0x01C2AC00 // PB00-PB01
#define TWI1_BASE_ADDRESS 0x01C2B000 // PB18-PB19
#define TWI2_BASE_ADDRESS 0x01C2B400 // PB20-PB21
#define TWI3_BASE_ADDRESS 0
#define TWI4_BASE_ADDRESS 0

#elif (ALLWINNER_H == 6)

#define TWI0_BASE_ADDRESS 0x05002000 // PD25-PD26
#define TWI1_BASE_ADDRESS 0x05002400 // PH05-PH06
#define TWI2_BASE_ADDRESS 0x05002800 // PD23-PD24
#define TWI3_BASE_ADDRESS 0x07081400 // PL00-PL01
#define TWI4_BASE_ADDRESS 0

#else // H5, H3 and H2+

#define TWI0_BASE_ADDRESS 0x01C2AC00 // PA11-PA12
#define TWI1_BASE_ADDRESS 0x01C2B000 // PA18-PA19
#define TWI2_BASE_ADDRESS 0x01C2B400 // PE12-PE13
#define TWI3_BASE_ADDRESS 0x01F02400 // PL00-PL01
#define TWI4_BASE_ADDRESS 0

#endif

typedef struct
{
  volatile uint32_t addr;   // 0x0000
  volatile uint32_t xaddr;  // 0x0004
  volatile uint32_t data;   // 0x0008
  volatile uint32_t cntr;   // 0x000C
  volatile uint32_t stat;   // 0x0010
  volatile uint32_t ccr;    // 0x0014
  volatile uint32_t srst;   // 0x0018
  volatile uint32_t efr;    // 0x001C
  volatile uint32_t lcr;    // 0x0020
} TWI_REG;

#define TWI_BUS_CLK (24000000 / 10)

#define TWI_BUS_CLK_M_POS 3
#define TWI_BUS_CLK_M_LEN 4
#define TWI_BUS_CLK_M_MSK ((1 << TWI_BUS_CLK_M_LEN) - 1)
#define TWI_BUS_CLK_N_POS 0
#define TWI_BUS_CLK_N_LEN 3
#define TWI_BUS_CLK_N_MSK ((1 << TWI_BUS_CLK_N_LEN) - 1)

class Allwinner_TWI
{
  private:

    TWI_REG *_twi;

  public:

    Allwinner_TWI(uint32_t num)
    : _twi(NULL)
    {
      static const uint32_t BASE[] = {TWI0_BASE_ADDRESS, TWI1_BASE_ADDRESS, TWI2_BASE_ADDRESS, TWI3_BASE_ADDRESS, TWI4_BASE_ADDRESS};
      if (num < sizeof(BASE) / sizeof(BASE[0]))
      {
        uint32_t base = BASE[num];
        if (base)
        {
          int fd = open("/dev/mem", O_RDWR);
          if (fd != -1)
          {
            unsigned int ps = sysconf(_SC_PAGESIZE);
            void *mm = mmap(0, ps, PROT_READ | PROT_WRITE, MAP_SHARED, fd, base & ~(ps - 1));
            if (mm == MAP_FAILED)
              fprintf(stderr, "[ERROR] mmap(0, %d, PROT_READ | PROT_WRITE, MAP_SHARED, %d, %08X) = %s\n",
                ps, fd, base & ~(ps - 1), strerror(errno));
            else
              _twi = (TWI_REG *)((uint8_t *)mm + (base & (ps - 1)));
            close(fd);
          }
        }
      }
    }

    virtual ~Allwinner_TWI(void)
    {
    }

    uint32_t getBusClock(void)
    {
      if (_twi == NULL)
        return 0;
      uint32_t clk_n = (_twi->ccr >> TWI_BUS_CLK_N_POS) & TWI_BUS_CLK_N_MSK;
      uint32_t clk_m = (_twi->ccr >> TWI_BUS_CLK_M_POS) & TWI_BUS_CLK_M_MSK;
      return (TWI_BUS_CLK >> clk_n) / (clk_m + 1);
    }

    void setBusClock(uint32_t f)
    {
      if ((_twi == NULL) || (f == 0))
        return;
      uint32_t clk_n = 0;
      uint32_t clk_m = 0;
      while (clk_n <= TWI_BUS_CLK_N_MSK)
      {
        clk_m = (TWI_BUS_CLK >> clk_n) / f;
        if (clk_m <= (TWI_BUS_CLK_M_MSK + 1))
          break;
        ++clk_n;
      }
      if ((clk_m > 0) && (clk_m <= (TWI_BUS_CLK_M_MSK + 1)))
        _twi->ccr = (_twi->ccr & ~((TWI_BUS_CLK_N_MSK << TWI_BUS_CLK_N_POS) | (TWI_BUS_CLK_M_MSK << TWI_BUS_CLK_M_POS)))
          | (clk_n << TWI_BUS_CLK_N_POS) | ((clk_m - 1) << TWI_BUS_CLK_M_POS);
    }
};

extern Allwinner_TWI TWI0[5];

#endif
