/*
  BlockingIO.h - Blocking I/O library

  Copyright (c) 2020 Sasapea's Lab. All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/
#ifndef __BLOCKINGIO_H
#define __BLOCKINGIO_H

#include <errno.h>
#include <stdint.h>
#include <stdbool.h>
#include <unistd.h>
#include "Stream.h"

class BlockingIO : public Stream
{
  private:

    unsigned long _timeout;
    int  _fd_r;
    int  _fd_w;
    int  _peek;
    int  _error;
    int  _refs;

  protected:

    virtual int fd()
    {
      return _fd_r;
    }

    virtual int available0(unsigned long timeout)
    {
      if (_peek >= 0)
        return 1;
      struct timeval tov = {(long)(timeout / 1000), (long)((timeout % 1000) * 1000)};
      fd_set fds[1];
      FD_ZERO(fds);
      FD_SET(_fd_r, fds);
      switch (select(_fd_r + 1, fds, NULL, NULL, &tov))
      {
        case -1:
          _error = 1;
        case 0:
          return 0;
      }
      return _error ? 0 : 1;
    }

    virtual int read(unsigned long timeout)
    {
      uint8_t b;
      if (_peek >= 0)
      {
        b = _peek;
        _peek = -1;
        return b;
      }
      if (available0(timeout))
      {
        if (::read(_fd_r, &b, sizeof(b)) == (ssize_t)sizeof(b))
          return b;
        _error = 1;
      }
      return -1;
    }

    virtual int availableForWrite(unsigned long timeout)
    {
      struct timeval tov = {(long)(timeout / 1000), (long)((timeout % 1000) * 1000)};
      fd_set fds[1];
      FD_ZERO(fds);
      FD_SET(_fd_w, fds);
      switch (select(_fd_w + 1, NULL, fds, NULL, &tov))
      {
        case -1:
          _error = 1;
        case 0:
          return 0;
      }
      return _error ? 0 : 1;
    }

    virtual size_t write(const uint8_t *buf, size_t len, unsigned long timeout)
    {
      if (availableForWrite(timeout))
      {
        if (::write(_fd_w, buf, len) == (ssize_t)len)
          return len;
        _error = 1;
      }
      return 0;
    }

    virtual void addRef(void)
    {
      ++_refs;
    }

    virtual int releaseRef(void)
    {
      return --_refs;
    }

    friend class EthernetClient;

  public:

    BlockingIO(void)
    : _fd_r(-1)
    , _fd_w(-1)
    , _peek(-1)
    , _error(0)
    , _refs(1)
    {
      _timeout = Stream::getTimeout();
      Stream::setTimeout(0);
    }

    virtual ~BlockingIO(void)
    {
    }

    virtual void setTimeout(unsigned long timeout)
    {
      _timeout = timeout;
    }

    virtual unsigned long getTimeout(void)
    {
      return _timeout;
    }

    virtual void begin(int fd = -1)
    {
      begin(fd, fd);
    }

    virtual void begin(int fd_r, int fd_w)
    {
      if ((_fd_r == -1) && (fd_r != -1) && (fd_w != -1))
      {
        int on = 1;
        if (ioctl(fd_r, FIONBIO, &on) == -1)
          fprintf(stderr, "[ERROR] ioctl(FIONBIO, %d) = %s\n", on, strerror(errno));
        if (fd_r != fd_w)
        {
          if (ioctl(fd_w, FIONBIO, &on) == -1)
            fprintf(stderr, "[ERROR] ioctl(FIONBIO, %d) = %s\n", on, strerror(errno));
        }
        _fd_r  = fd_r;
        _fd_w  = fd_w;
        _peek  = -1;
        _error = 0;
      }
    }

    virtual void end(bool closing = true)
    {
      if (_fd_r != -1)
      {
        if (closing)
        {
          if (_fd_r > 2)
            close(_fd_r);
          if (_fd_w > 2)
            close(_fd_w);
        }
        _fd_r  = -1;
        _fd_w  = -1;
        _peek  = -1;
        _error = 0;
      }
    }

    virtual void flush(void)
    {
    }

    virtual int available(void)
    {
      return available(0);
    }

    virtual int available(unsigned long timeout)
    {
      return available0(timeout);
    }

    virtual int peek(void)
    {
      if (_peek < 0)
        _peek = read(0);
      return _peek;
    }

    virtual int read(void)
    {
      return read(getTimeout());
    }

    virtual int read(uint8_t *buf, size_t size)
    {
      for (size_t i = 0; i < size; ++i)
      {
        int c = read();
        if (c < 0)
          return i;
        *buf++ = (uint8_t)c;
      }
      return size;
    }

    virtual int availableForWrite(void)
    {
      return availableForWrite(0);
    }

    virtual size_t write(uint8_t b)
    {
      return write(&b, sizeof(b));
    }

    virtual size_t write(const uint8_t *buf, size_t size)
    {
      return write(buf, size, getTimeout());
    }

    virtual bool opened()
    {
      return _fd_r != -1;
    }

    virtual bool connected(void)
    {
      return opened() && !_error;
    }
};

#endif
