/*
  EthernetClient.h - TCP Client Class

  Copyright (c) 2020 Sasapea's Lab. All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/
#ifndef __ETHERNETCLIENT_H
#define __ETHERNETCLIENT_H

#include <stdio.h>
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include "Client.h"
#include "IPAddress.h"
#include "BlockingIO.h"

class EthernetClient : public Client
{
  private:

    BlockingIO *_socket;
    IPAddress _remoteIP;
    String    _remoteHost;
    uint16_t  _remotePort;
    uint16_t  _localPort;
    unsigned long _timeout;

  protected:

    struct addrinfo *getaddrinfo(const char *node, uint16_t port, int socktype, int flags = 0)
    {
      struct addrinfo hints, *res;
      char service[16];
      snprintf(service, sizeof(service), "%u", port);
      memset(&hints, 0, sizeof(hints));
      hints.ai_family   = AF_UNSPEC;
      hints.ai_socktype = socktype;
      hints.ai_flags    = flags;
      return ::getaddrinfo(node, service, &hints, &res) == 0 ? res : NULL;
    }

    bool open(int fd)
    {
      char host[NI_MAXHOST], service[NI_MAXSERV];
      struct sockaddr_storage addr;
      socklen_t len = sizeof(addr);
      if (getpeername(fd, (struct sockaddr *)&addr, &len) != 0)
        return false;
      if (getnameinfo((struct sockaddr *)&addr, len, host, sizeof(host),
          service, sizeof(service), NI_NUMERICHOST | NI_NUMERICSERV) != 0)
        return false;
      _remoteIP.fromString(host); // IPv6 Not Supported.
      _remoteHost = host;
      _remotePort = atoi(service);
      len = sizeof(addr);
      if (getsockname(fd, (struct sockaddr *)&addr, &len) != 0)
        return false;
      if (getnameinfo((struct sockaddr *)&addr, len, host, sizeof(host),
          service, sizeof(service), NI_NUMERICHOST | NI_NUMERICSERV) != 0)
        return false;
       _localPort = atoi(service);
       _socket->begin(fd);
       return true;
    }

    void release(void)
    {
      _socket->releaseRef();
      _socket = new BlockingIO();
    }

    friend class EthernetServer;

  public:

    EthernetClient(const EthernetClient &obj)
    {
      _socket     = obj._socket;
      _remoteIP   = obj._remoteIP;
      _remoteHost = obj._remoteHost;
      _remotePort = obj._remotePort;
      _localPort  = obj._localPort;
      _timeout    = obj._timeout;
      _socket->addRef();
    }

    EthernetClient(int fd = -1)
    : _remoteIP((uint32_t)0)
    , _remotePort(0)
    , _localPort(0)
    , _timeout(3000)
    {
      Stream::setTimeout(0);
      _socket = new BlockingIO();
      _socket->begin(fd);
    }

    virtual ~EthernetClient(void)
    {
      if (_socket->releaseRef() == 0)
        delete _socket;
    }

    void setTimeout(unsigned long timeout)
    {
      _socket->setTimeout(timeout);
    }

    unsigned long getTimeout(void)
    {
      return _socket->getTimeout();
    }

    void setConnectionTimeout(unsigned long timeout)
    {
      _timeout = timeout;
    }

    int connect(IPAddress ip, uint16_t port)
    {
      char host[32];
      snprintf(host, sizeof(host), "%u.%u.%u.%u", ip[0], ip[1], ip[2], ip[3]);
      return connect(host, port);
    }

    int connect(const char *node, uint16_t port)
    {
      struct addrinfo *ai, *res = getaddrinfo(node, port, SOCK_STREAM);
      stop();
      if (res)
      {
        for (ai = res; ai; ai = ai->ai_next)
        {
          int fd = socket(ai->ai_family, ai->ai_socktype | SOCK_NONBLOCK, ai->ai_protocol);
          if (fd == -1)
            continue;
          if (::connect(fd, ai->ai_addr, ai->ai_addrlen) == 0)
          {
            if (open(fd))
              break;
          }
          else if (errno == EINPROGRESS)
          {
            struct timeval tov = {(long)(_timeout / 1000), (long)((_timeout % 1000) * 1000)};
            fd_set fds[1];
            FD_ZERO(fds);
            FD_SET(fd, fds);
            if (select(fd + 1, NULL, fds, NULL, &tov) > 0)
            {
              if (open(fd))
                break;
            }
          }
          close(fd);
        }
        freeaddrinfo(res);
      }
      return _socket->opened();
    }

    void stop(void)
    {
      _socket->end();
    }

    void flush(void)
    {
      _socket->flush();
    }

    int available(void)
    {
      return _socket->available();
    }

    int available(unsigned long timeout)
    {
      return _socket->available(timeout);
    }

    int peek(void)
    {
      return _socket->peek();
    }

    int read(void)
    {
      return _socket->read();
    }

    int read(uint8_t *buf, size_t size)
    {
      return _socket->read(buf, size);
    }

    int availableForWrite(void)
    {
      return _socket->availableForWrite();
    }

    size_t write(uint8_t b)
    {
      return _socket->write(b);
    }

    size_t write(const uint8_t *buf, size_t size)
    {
      return _socket->write(buf, size);
    }

    uint8_t connected()
    {
      return _socket->connected();
    }

    IPAddress remoteIP()
    {
      return _remoteIP;
    }

    String remoteHost()
    {
      return _remoteHost;
    }

    uint16_t remotePort()
    {
      return _remotePort;
    }

    uint16_t localPort()
    {
      return _localPort;
    }

    operator bool()
    {
      return _socket->opened();
    }
};

#endif
