/*
  EthernetServer.h - Tcp Server Class

  Copyright (c) 2020 Sasapea's Lab. All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/
#ifndef __ETHERNETSERVER_H
#define __ETHERNETSERVER_H

#include <string.h>
#include <netdb.h>
#include <netinet/tcp.h>
#include "Server.h"
#include "EthernetClient.h"

#ifndef TCPSERVER_BACKLOG
#define TCPSERVER_BACKLOG 5
#endif

class EthernetServer : public Server
{
  private:

    EthernetClient _remote;
    uint16_t _port;
    int      _sockfds[2];
    size_t   _sockcnt;
    int      _sockmax;

    struct addrinfo *getaddrinfo(const char *node, uint16_t port, int socktype, int flags = 0)
    {
      struct addrinfo hints, *res;
      char service[16];
      snprintf(service, sizeof(service), "%u", port);
      memset(&hints, 0, sizeof(hints));
      hints.ai_family   = AF_UNSPEC;
      hints.ai_socktype = socktype;
      hints.ai_flags    = flags;
      return ::getaddrinfo(node, service, &hints, &res) == 0 ? res : NULL;
    }

    void listen(void)
    {
      struct addrinfo *ai, *res = getaddrinfo(NULL, _port, SOCK_STREAM, AI_PASSIVE);
      _sockcnt = 0;
      _sockmax = -1;
      if (res)
      {
        for (ai = res; ai && (_sockcnt < sizeof(_sockfds) / sizeof(_sockfds[0])); ai = ai->ai_next)
        {
          int fd, on = 1;
          fd = socket(ai->ai_family, ai->ai_socktype | SOCK_NONBLOCK, ai->ai_protocol);
          if (fd == -1)
            continue;
          switch (ai->ai_family)
          {
            case AF_INET6:
              if (setsockopt(fd, IPPROTO_IPV6, IPV6_V6ONLY, &on, sizeof(on)) == -1)
                fprintf(stderr, "[ERROR] setsockopt(IPPROTO_IPV6, IPV6_V6ONLY, %d) = %s\n", on, strerror(errno));
            case AF_INET:
              if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on)) == -1)
                fprintf(stderr, "[ERROR] setsockopt(SOL_SOCKET, SO_REUSEADDR, %d) = %s\n", on, strerror(errno));
              break;
          }
          if (::bind(fd, ai->ai_addr, ai->ai_addrlen) == 0)
          {
            if (::listen(fd, TCPSERVER_BACKLOG) == 0)
            {
              _sockfds[_sockcnt++] = fd;
              if (_sockmax < fd)
                _sockmax = fd;
              continue;
            }
          }
          close(fd);
        }
        freeaddrinfo(res);
      }
    }

    bool accept0(unsigned long timeout)
    {
      while (!_remote.connected())
      {
        struct timeval tov = {(long)(timeout / 1000), (long)((timeout % 1000) * 1000)};
        fd_set fds[sizeof(_sockfds) / sizeof(_sockfds[0])];
       _remote->stop();
        FD_ZERO(fds);
        for (size_t i = 0; i < _sockcnt; ++i)
          FD_SET(_sockfds[i], fds);
        if (select(_sockmax + 1, fds, NULL, NULL, &tov) <= 0)
          break;
        for (size_t i = 0; i < _sockcnt; ++i)
        {
          if (!FD_ISSET(_sockfds[i], fds))
            continue;
          struct sockaddr_storage addr;
          socklen_t len = sizeof(addr);
          int fd = ::accept(_sockfds[i], (struct sockaddr *)&addr, &len);
          if (fd == -1)
            continue;
          if (_remote.open(fd))
            break;
          close(fd);
        }
      }
      return _remote;
    }

  public:

    EthernetServer(uint16_t port)
    : _port(port)
    , _sockcnt(0)
    {
    }

    virtual ~EthernetServer(void)
    {
    }

    void begin()
    {
      if (_sockcnt == 0)
        listen();
    }

    void end(void)
    {
      _remote.stop();
      for (size_t i = 0; i < _sockcnt; ++i)
        close(_sockfds[i]);
      _sockcnt = 0;
    }

    EthernetClient available(unsigned long timeout = 0)
    {
      return accept0(timeout) && _remote.available() ? _remote : EthernetClient();
    }

    EthernetClient accept(unsigned long timeout = 0)
    {
      if (!accept0(timeout))
        return EthernetClient();
      EthernetClient client(_remote);
      _remote.release();
      return client;
    }

    void flush(void)
    {
      _remote.flush();
    }

    int availableForWrite(void)
    {
      return _remote.availableForWrite();
    }

    size_t write(uint8_t b)
    {
      return _remote.write(b);
    }

    operator bool()
    {
      return _remote == true;
    }
};

#endif
