/*
  EthernetUdp.h - Udp Class

  Copyright (c) 2020 Sasapea's Lab. All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/
#ifndef __ETHERNETUDP_H
#define __ETHERNETUDP_H

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include "Udp.h"

#define UDP_PACKET_MAX_SIZE 2048

class EthernetUdp : public UDP
{
  private:

    uint16_t  _port;
    uint8_t   _buffer[UDP_PACKET_MAX_SIZE];
    uint16_t  _offset;
    uint16_t  _count;
    IPAddress _remoteIP;
    String    _remoteHost;
    uint16_t  _remotePort;
    struct  sockaddr_storage _addr;
    socklen_t _addrlen;
    int       _sockafs[2];
    int       _sockfds[2];
    size_t    _sockcnt;
    int       _sockmax;
    int       _sendfd;

  protected:

    struct addrinfo *getaddrinfo(const char *node, uint16_t port, int socktype, int flags = 0)
    {
      struct addrinfo hints, *res;
      char service[16];
      snprintf(service, sizeof(service), "%u", port);
      memset(&hints, 0, sizeof(hints));
      hints.ai_family   = AF_UNSPEC;
      hints.ai_socktype = socktype;
      hints.ai_flags    = flags;
      return ::getaddrinfo(node, service, &hints, &res) == 0 ? res : NULL;
    }

    uint8_t begin(const char *mcaddr, const char *mcaddr6, uint16_t port)
    {
      struct addrinfo *ai, *res = getaddrinfo(NULL, port, SOCK_DGRAM, AI_PASSIVE);
      _sockcnt = 0;
      _sockmax = -1;
      if (res)
      {
        for (ai = res; ai && (_sockcnt < sizeof(_sockfds) / sizeof(_sockfds[0])); ai = ai->ai_next)
        {
          int fd, on = 1;
          fd = socket(ai->ai_family, ai->ai_socktype | SOCK_NONBLOCK, ai->ai_protocol);
          if (fd == -1)
            continue;
          switch (ai->ai_family)
          {
            case AF_INET6:
              if (setsockopt(fd, IPPROTO_IPV6, IPV6_V6ONLY, &on, sizeof(on)) == -1)
                fprintf(stderr, "[ERROR] setsockopt(IPPROTO_IPV6, IPV6_V6ONLY, %d) = %s\n", on, strerror(errno));
            case AF_INET:
              if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on)) == -1)
                fprintf(stderr, "[ERROR] setsockopt(SOL_SOCKET, SO_REUSEADDR, %d) = %s\n", on, strerror(errno));
              if (setsockopt(fd, SOL_SOCKET, SO_BROADCAST, &on, sizeof(on)) == -1)
                fprintf(stderr, "[ERROR] setsockopt(SOL_SOCKET, SO_BROADCAST, %d) = %s\n", on, strerror(errno));
              break;
          }
          if (bind(fd, ai->ai_addr, ai->ai_addrlen) == 0)
          {
            switch (ai->ai_family)
            {
              case AF_INET6:
                if (mcaddr6)
                {
                  struct ipv6_mreq mreq6;
                  memset(&mreq6, 0, sizeof(mreq6));
                  if (inet_pton(ai->ai_family, mcaddr6, &mreq6.ipv6mr_multiaddr) == 1)
                  {
                    if (setsockopt(fd, IPPROTO_IPV6, IPV6_ADD_MEMBERSHIP, &mreq6, sizeof(mreq6)) == -1)
                      fprintf(stderr, "[ERROR] setsockopt(IPPROTO_IPV6, IPV6_ADD_MEMBERSHIP, \"%s\") = %s\n", mcaddr6, strerror(errno));
                  }
                }
                break;
              case AF_INET:
                if (mcaddr)
                {
                  struct ip_mreq mreq;
                  memset(&mreq, 0, sizeof(mreq));
                  if (inet_pton(ai->ai_family, mcaddr, &mreq.imr_multiaddr) == 1)
                  {
                    if (setsockopt(fd, IPPROTO_IP, IP_ADD_MEMBERSHIP, &mreq, sizeof(mreq)) == -1)
                      fprintf(stderr, "[ERROR] setsockopt(IPPROTO_IP, IP_ADD_MEMBERSHIP, \"%s\") = %s\n", mcaddr, strerror(errno));
                  }
                }
                break;
            }
            _sockafs[_sockcnt  ] = ai->ai_family;
            _sockfds[_sockcnt++] = fd;
            if (_sockmax < fd)
              _sockmax = fd;
            continue;
          }
          close(fd);
        }
        freeaddrinfo(res);
      }
      return _sockcnt > 0;
    }

  public:

    EthernetUdp(void)
    : _port(0)
    , _offset(0)
    , _count(0)
    , _remoteIP((uint32_t)0)
    , _remotePort(0)
    , _sockcnt(0)
    , _sockmax(-1)
    , _sendfd(-1)
    {
    }

    virtual ~EthernetUdp(void)
    {
    }

    uint8_t begin(uint16_t port)
    {
      return beginMulticast(NULL, port);
    }

    uint8_t beginMulticast(IPAddress ip, uint16_t port)
    {
      char addr[32];
      snprintf(addr, sizeof(addr), "%u.%u.%u.%u", ip[0], ip[1], ip[2], ip[3]);
      return beginMulticast(addr, port);
    }

    uint8_t beginMulticast(const char *mcaddr, uint16_t port)
    {
      return beginMulticast(mcaddr, mcaddr, port);
    }

    uint8_t beginMulticast(const char *mcaddr, const char *mcaddr6, uint16_t port)
    {
      return _sockcnt ? 0 : begin(mcaddr, mcaddr6, port);
    }

    void stop(void)
    {
      for (size_t i = 0; i < _sockcnt; ++i)
      {
        close(_sockfds[i]);
        _sockfds[i] = -1;
      }
      _sockcnt = 0;
    }

    int beginPacket(IPAddress ip, uint16_t port)
    {
      char addr[32];
      snprintf(addr, sizeof(addr), "%u.%u.%u.%u", ip[0], ip[1], ip[2], ip[3]);
      return beginPacket(addr, port);
    }

    int beginPacket(const char *node, uint16_t port)
    {
      struct addrinfo *ai, *res = getaddrinfo(node, port, SOCK_DGRAM);
      _count  = 0;
      _offset = 0;
      _sendfd = -1;
      if (res)
      {
        for (ai = res; ai && (_sendfd == -1); ai = ai->ai_next)
        {
          for (size_t i = 0; i < _sockcnt; ++i)
          {
            if (_sockafs[i] == ai->ai_family)
            {
              _sendfd  = _sockfds[i];
              _addrlen = ai->ai_addrlen;
              memcpy(&_addr, ai->ai_addr, ai->ai_addrlen);
              break;
            }
          }
        }
        freeaddrinfo(res);
      }
      return _sendfd != -1 ? 1 : 0;
    }

    int endPacket(void)
    {
      return endPacket(1000);
    }

    int endPacket(unsigned long timeout)
    {
      int rv = 0;
      if ((_sendfd != -1) && (_offset > 0))
      {
        struct timeval tov = {(long)(timeout / 1000), (long)((timeout % 1000) * 1000)};
        fd_set fds[1];
        FD_ZERO(fds);
        FD_SET(_sendfd, fds);
        if (select(_sendfd + 1, NULL, fds, NULL, &tov) > 0)
        {
          ssize_t n = sendto(_sendfd, _buffer, _offset, 0, (struct sockaddr *)&_addr, _addrlen);
          if (n >= 0)
            rv = (n == (ssize_t)_offset);
          else if (n == -1)
          {
            char host[NI_MAXHOST], service[NI_MAXSERV];
            if (getnameinfo((struct sockaddr *)&_addr, _addrlen, host, sizeof(host),
                service, sizeof(service), NI_NUMERICHOST | NI_NUMERICSERV) == 0)
              fprintf(stderr, "[ERROR] sendto(%d, \"%s\":%s) = %d (%s)\n", _offset, host, service, n, strerror(errno));
          }
        }
      }
      _sendfd = -1;
      _offset = 0;
      return rv;
    }

    size_t write(uint8_t b)
    {
      if ((_sendfd != -1) && (_offset < sizeof(_buffer)))
      {
        _buffer[_offset++] = b;
        return 1;
      }
      return 0;
    }

    size_t write(const uint8_t *buffer, size_t size)
    {
      if (_sendfd == -1)
        return 0;
      if (size > sizeof(_buffer) - _offset)
        size = sizeof(_buffer) - _offset;
      memcpy(_buffer, buffer, size);
      _offset += size;
      return size;
    }

    int parsePacket(void)
    {
      return parsePacket(0);
    }

    int parsePacket(unsigned long timeout)
    {
      struct timeval tov = {(long)(timeout / 1000), (long)((timeout % 1000) * 1000)};
      fd_set fds[sizeof(_sockfds) / sizeof(_sockfds[0])];
      FD_ZERO(fds);
      for (size_t i = 0; i < _sockcnt; ++i)
        FD_SET(_sockfds[i], fds);
      _offset = 0;
      _count  = 0;
      if (select(_sockmax + 1, fds, NULL, NULL, &tov) > 0)
      {
        for (size_t i = 0; i < _sockcnt; ++i)
        {
          if (!FD_ISSET(_sockfds[i], fds))
            continue;
          struct sockaddr_storage addr;
          socklen_t addrlen = sizeof(addr);
          ssize_t n = recvfrom(_sockfds[i], _buffer, sizeof(_buffer), 0, (struct sockaddr *)&addr, &addrlen);
          if (n > 0)
          {
            char host[NI_MAXHOST], service[NI_MAXSERV];
            if (getnameinfo((struct sockaddr *)&addr, addrlen,
                  host, sizeof(host), service, sizeof(service), NI_NUMERICHOST | NI_NUMERICSERV) == 0)
            {
              _remoteIP.fromString(host); // IPv6 Not Supported.
              _remoteHost = host;
              _remotePort = atoi(service);
              _count  = n;
              break;
            }
          }
        }
      }
      return _count;
    }

    int available(void)
    {
      return _count - _offset;
    }

    int read(void)
    {
      return available() ? _buffer[_offset++] : -1;
    }

    int read(unsigned char *buffer, size_t size)
    {
      if (size > (size_t)available())
        size = available();
      memcpy(buffer, _buffer, size);
      _offset += size;
      return size;
    }

    int read(char *buffer, size_t size)
    {
      return read((unsigned char *)buffer, size);
    }

    int peek(void)
    {
      return available() ? _buffer[_offset] : -1;
    }

    void flush(void)
    {
    }

    IPAddress remoteIP(void)
    {
      return _remoteIP;
    }

    String remoteHost(void)
    {
      return _remoteHost;
    }

    uint16_t remotePort(void)
    {
      return _remotePort;
    }
};

#endif
