/*
  HardwareSerial.h - Hardware serial library for Wiring

  Copyright (c) 2020 Sasapea's Lab. All right reserved.

  System Setting:

    [Oraneg Pi Zeo][/boot/armbianEnv.txt]
    overlays=... uart1 uart2 ...
    param_uart1_rtscts=0
    param_uart2_rtscts=0

    /dev/ttyS1
    /dev/ttyS2
    /dev/ttyS3
    /dev/ttyACM0
    /dev/ttyACM1
    /dev/ttyACM2

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/
#ifndef __HARDWARESERIAL_H
#define __HARDWARESERIAL_H

#include <stdio.h>
#include <string.h>
#include <error.h>
#include <termios.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include "BlockingIO.h"

#define SERIAL_5N1 0
#define SERIAL_6N1 1
#define SERIAL_7N1 2
#define SERIAL_8N1 3
#define SERIAL_5N2 4
#define SERIAL_6N2 5
#define SERIAL_7N2 6
#define SERIAL_8N2 7
#define SERIAL_5E1 8
#define SERIAL_6E1 9
#define SERIAL_7E1 10
#define SERIAL_8E1 11
#define SERIAL_5E2 12
#define SERIAL_6E2 13
#define SERIAL_7E2 14
#define SERIAL_8E2 15
#define SERIAL_5O1 16
#define SERIAL_6O1 17
#define SERIAL_7O1 18
#define SERIAL_8O1 19
#define SERIAL_5O2 20
#define SERIAL_6O2 21
#define SERIAL_7O2 22
#define SERIAL_8O2 23

class HardwareSerial : public BlockingIO
{
  private:

    struct termios _tcattr;
    const char *_name;

  public:

    HardwareSerial(void)
    : HardwareSerial(NULL)
    {
      memset(&_tcattr, 0, sizeof(_tcattr));
    }

    HardwareSerial(const char *name)
    : _name(name)
    {
      memset(&_tcattr, 0, sizeof(_tcattr));
    }

    virtual ~HardwareSerial(void)
    {
    }

    void begin(int fd_r, int fd_w)
    {
      BlockingIO::begin(fd_r, fd_w);
    }

    void begin(unsigned long baud, uint8_t config = SERIAL_8N1)
    {
      int fd;
      if (opened())
        return;
      if (_name == NULL)
      {
        begin(0, 1); // stdin and stdout
        return;
      }
      if ((fd = open(_name, O_RDWR | O_NOCTTY | O_NDELAY | O_NONBLOCK)) == -1)
      {
        fprintf(stderr, "[ERROR] open(%s) = %s\n", _name, strerror(errno));
        return;
      }
      tcflag_t speed, frame;
      struct termios tc;
      tcgetattr(fd, &tc);
      _tcattr = tc;
      cfmakeraw(&tc);
      speed = cfgetospeed(&tc);
      frame = tc.c_cflag & (CSIZE | CSTOPB | PARENB | PARODD);
      switch (baud)
      {
        case      50: speed =      B50; break;
        case      75: speed =      B75; break;
        case     110: speed =     B110; break;
        case     134: speed =     B134; break;
        case     150: speed =     B150; break;
        case     200: speed =     B200; break;
        case     300: speed =     B300; break;
        case     600: speed =     B600; break;
        case    1200: speed =    B1200; break;
        case    1800: speed =    B1800; break;
        case    2400: speed =    B2400; break;
        case    4800: speed =    B4800; break;
        case    9600: speed =    B9600; break;
        case   19200: speed =   B19200; break;
        case   38400: speed =   B38400; break;
        case   57600: speed =   B57600; break;
        case  115200: speed =  B115200; break;
        case  230400: speed =  B230400; break;
        case  460800: speed =  B460800; break;
        case  500000: speed =  B500000; break;
        case  576000: speed =  B576000; break;
        case  921600: speed =  B921600; break;
        case 1000000: speed = B1000000; break;
        case 1152000: speed = B1152000; break;
        case 1500000: speed = B1500000; break;
        case 2000000: speed = B2000000; break;
        case 2500000: speed = B2500000; break;
        case 3000000: speed = B3000000; break;
        case 3500000: speed = B3500000; break;
        case 4000000: speed = B4000000; break;
      }
      switch (config)
      {
        case SERIAL_5N1: frame = CS5; break;
        case SERIAL_6N1: frame = CS6; break;
        case SERIAL_7N1: frame = CS7; break;
        case SERIAL_8N1: frame = CS8; break;
        case SERIAL_5N2: frame = CS5 | CSTOPB; break;
        case SERIAL_6N2: frame = CS6 | CSTOPB; break;
        case SERIAL_7N2: frame = CS7 | CSTOPB; break;
        case SERIAL_8N2: frame = CS8 | CSTOPB; break;
        case SERIAL_5E1: frame = CS5 | PARENB; break;
        case SERIAL_6E1: frame = CS6 | PARENB; break;
        case SERIAL_7E1: frame = CS7 | PARENB; break;
        case SERIAL_8E1: frame = CS8 | PARENB; break;
        case SERIAL_5E2: frame = CS5 | CSTOPB | PARENB; break;
        case SERIAL_6E2: frame = CS6 | CSTOPB | PARENB; break;
        case SERIAL_7E2: frame = CS7 | CSTOPB | PARENB; break;
        case SERIAL_8E2: frame = CS8 | CSTOPB | PARENB; break;
        case SERIAL_5O1: frame = CS5 | PARENB | PARODD; break;
        case SERIAL_6O1: frame = CS6 | PARENB | PARODD; break;
        case SERIAL_7O1: frame = CS7 | PARENB | PARODD; break;
        case SERIAL_8O1: frame = CS8 | PARENB | PARODD; break;
        case SERIAL_5O2: frame = CS5 | CSTOPB | PARENB | PARODD; break;
        case SERIAL_6O2: frame = CS6 | CSTOPB | PARENB | PARODD; break;
        case SERIAL_7O2: frame = CS7 | CSTOPB | PARENB | PARODD; break;
        case SERIAL_8O2: frame = CS8 | CSTOPB | PARENB | PARODD; break;
      }
      cfsetspeed(&tc, speed);
      tc.c_cflag = (tc.c_cflag & ~(CSIZE | CSTOPB | PARENB | PARODD)) | frame;
      tc.c_iflag &= ~IXON;
      tc.c_oflag &= ~OPOST;
      tcflush(fd, TCIOFLUSH);
      tcsetattr(fd, TCSANOW, &tc);
      BlockingIO::begin(fd);
    }

    void end(void)
    {
      if (opened())
      {
        if (_tcattr.c_cflag)
          tcsetattr(fd(), TCSANOW, &_tcattr);
        BlockingIO::end();
      }
    }

    operator bool(void)
    {
      return opened();
    }
};

extern HardwareSerial Serial;
extern HardwareSerial Serial1;
extern HardwareSerial Serial2;
extern HardwareSerial Serial3;
extern HardwareSerial SerialUSB;
extern HardwareSerial SerialUSB1;
extern HardwareSerial SerialUSB2;
extern HardwareSerial SerialUSB3;
extern HardwareSerial SerialACM;
extern HardwareSerial SerialACM1;
extern HardwareSerial SerialACM2;
extern HardwareSerial SerialACM3;
extern HardwareSerial SerialSTD;

extern void serialEventRun(void) __attribute__((weak));

#endif
