/*
  PWMCTRL.h - PWM Control class
  Caution: SIGUSR1 signal setup required.

  Copyright (c) 2020 Sasapea's Lab. All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/
#ifndef __PWMCTRL_H
#define __PWMCTRL_H

#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>
#include "ThreadTimer.h"
#include "ThreadPWM.h"
#include "Allwinner.h"

extern unsigned long PWMCTRL_timeup(void *args, unsigned long us);

struct pwm_ctrl
{
  uint16_t pin;
  PWMBase *pwm;
  ThreadTimer *timer;
  struct pwm_ctrl *next;
};

class PWMCTRL
{
  private:

    struct pwm_ctrl *_ctrl_list;

    void add(struct pwm_ctrl *ctrl)
    {
      ctrl->next = _ctrl_list;
      _ctrl_list = ctrl;
    }

    struct pwm_ctrl *get(uint16_t pin, bool remove = false)
    {
      for (struct pwm_ctrl **p = &_ctrl_list; *p; p = &(*p)->next)
      {
        if ((*p)->pin == pin)
        {
          struct pwm_ctrl *ctrl = *p;
          if (remove)
            *p = (*p)->next;
          return ctrl;
        }
      }
      return NULL;
    }

  public:

    PWMCTRL(void)
    : _ctrl_list(NULL)
    {
    }

    virtual ~PWMCTRL(void)
    {
    }

    void begin(uint16_t pin, uint32_t frequency, uint32_t duty, uint32_t duration = 0)
    {
      struct pwm_ctrl *ctrl = get(pin);
      if (ctrl == NULL)
      {
        ctrl = (struct pwm_ctrl *)malloc(sizeof(*ctrl));
        if (ctrl)
        {
          ctrl->pin = pin;
          if (pin == PWM0_PORT)
            ctrl->pwm = (PWMBase *)&PWM0[0];
          else if (pin == PWM1_PORT)
            ctrl->pwm = (PWMBase *)&PWM0[1];
          else
            ctrl->pwm = (PWMBase *)new ThreadPWM(pin);
          ctrl->timer = new ThreadTimer();
          add(ctrl);
        }
      }
      if (ctrl)
      {
        ctrl->timer->end();
        ctrl->pwm->polarity(PWM_ACTIVE_HIGH);
        ctrl->pwm->frequency(frequency);
        ctrl->pwm->duty(duty);
        ctrl->pwm->begin();
        if (duration)
          ctrl->timer->begin(duration, PWMCTRL_timeup, ctrl);
      }
    }

    void end(uint16_t pin)
    {
      struct pwm_ctrl *ctrl = get(pin, true);
      if (ctrl)
      {
        ctrl->timer->end();
        ctrl->pwm->end();
        delete ctrl->timer;
        if ((pin != PWM0_PORT) && (pin != PWM1_PORT))
          delete ctrl->pwm;
        free(ctrl);
      }
    }

    void polarity(uint16_t pin, int value)
    {
      struct pwm_ctrl *ctrl = get(pin);
      if (ctrl)
        ctrl->pwm->polarity(value);
    }

    void frequency(uint16_t pin, uint32_t value)
    {
      struct pwm_ctrl *ctrl = get(pin);
      if (ctrl)
        ctrl->pwm->frequency(value);
    }

    void duty(uint16_t pin, uint32_t value)
    {
      struct pwm_ctrl *ctrl = get(pin);
      if (ctrl)
        ctrl->pwm->duty(value);
    }
};

extern PWMCTRL PWM;

#endif
