/*
  SPI library for Linux

  Copyright (c) 2020 Sasapea's Lab. All right reserved.

  System Setting:

    [Oraneg Pi Zero][/boot/armbianEnv.txt]
    overlays=... [spi-add-cs1] spi-spidev ...
    param_spidev_spi_bus=1
    param_spidev_max_freq=1000000 (optinal. default: 1000000, Range: 3000 - 100000000)
    [param_spidev_spi_cs=0 or 1 (optional. default: 0)]

    /dev/spidev${param_spidev_spi_bus}.${param_spidev_spi_cs}

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/
#ifndef __SPI_H
#define __SPI_H

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdint.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/types.h>
#include <linux/spi/spidev.h>

#ifndef LSBFIRST
#define LSBFIRST 0
#endif

#ifndef MSBFIRST
#define MSBFIRST 1
#endif

#define SPI_MODE0 SPI_MODE_0
#define SPI_MODE1 SPI_MODE_1
#define SPI_MODE2 SPI_MODE_2
#define SPI_MODE3 SPI_MODE_3

class SPISettings
{
  protected:
    struct spi_ioc_transfer _transfer;
    uint8_t  _bitOrder;
    uint8_t  _dataMode;

    void copy(SPISettings settings)
    {
      _transfer = settings._transfer;
      _bitOrder = settings._bitOrder;
      _dataMode = settings._dataMode;
    }

  public:
    SPISettings(uint32_t speed_hz = 4000000, uint8_t bitOrder = MSBFIRST, uint8_t dataMode = SPI_MODE0, uint8_t delay_usecs = 0, uint8_t bits_per_word = 8, uint8_t cs_change = 0)
    : _bitOrder(!bitOrder)
    , _dataMode(dataMode)
    {
      memset(&_transfer, 0, sizeof(_transfer));
      _transfer.speed_hz      = speed_hz;
      _transfer.delay_usecs   = delay_usecs;
      _transfer.bits_per_word = bits_per_word;
      _transfer.cs_change     = cs_change;
    }
};

class SPIClass : SPISettings
{
  private:

    const char *_name;
    int _fd;

    void setup(void)
    {
      if (ioctl(_fd, SPI_IOC_WR_MODE, &_dataMode) == -1)
        fprintf(stderr, "[ERROR] ioctl(SPI_IOC_WR_MODE, %d) = %s\n", _dataMode, strerror(errno));
      if (ioctl(_fd, SPI_IOC_WR_LSB_FIRST, &_bitOrder) == -1)
        fprintf(stderr, "[ERROR] ioctl(SPI_IOC_WR_LSB_FIRST, %d) = %s\n", _bitOrder, strerror(errno));
    }

  public:

    SPIClass(const char *name)
    : _name(name)
    , _fd(-1)
    {
    }

    void beginTransaction(SPISettings settings)
    {
      copy(settings);
      setup();
    }

    void endTransaction(void)
    {
    }

    void begin()
    {
      if (_fd == -1)
      {
        if ((_fd = open(_name, O_RDWR)) != -1)
          setup();
      }
    }

    void end()
    {
      if (_fd != -1)
      {
        close(_fd);
        _fd = -1;
      }
    }

    void transfer(void *buf, size_t count)
    {
      _transfer.tx_buf = (int)buf;
      _transfer.rx_buf = (int)buf;
      _transfer.len    = count;
      if (ioctl(_fd, SPI_IOC_MESSAGE(1), &_transfer) == -1)
        fprintf(stderr, "[ERROR] SPIClass::transfer() ioctl(SPI_IOC_MESSAGE)\n");
    }

    uint8_t transfer(uint8_t data)
    {
      transfer(&data, sizeof(data));
      return data;
    }

    uint16_t transfer16(uint16_t data)
    {
      uint8_t b[2];
      if (_bitOrder == MSBFIRST)
      {
        b[0] = data >> 8;
        b[1] = data & 0xFF;
        transfer(b, sizeof(b));
        data = (b[0] << 8) | b[1];
      }
      else
      {
        b[0] = data & 0xFF;
        b[1] = data >> 8;
        transfer(b, sizeof(b));
        data = ((b[1]) << 8) | b[0];
      }
      return data;
    }
};

extern SPIClass SPI;
extern SPIClass SPI1;
extern SPIClass SPI2;
extern SPIClass SPI3;

#endif
