/*
  ThreadBase.h - Thread Base Class
  Caution: SIGUSR1 signal setup required.

  Copyright (c) 2020 Sasapea's Lab. All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/
#ifndef __THREADBASE_H
#define __THREADBASE_H

#include <stdbool.h>
#include <signal.h>
#include <pthread.h>

extern void *ThreadBase_run(void *obj);

class ThreadBase
{
  private:

    pthread_t _thread;
    bool      _running;
    bool      _detach;

  protected:

    virtual bool running()
    {
      return _running;
    }

    virtual void detach(void)
    {
      pthread_detach(_thread);
      _detach = true;
    }

    virtual void set_thread_priority(int new_policy, int new_priority)
    {
      int policy;
      struct sched_param param;
      pthread_getschedparam(_thread, &policy, &param);
      param.sched_priority = new_priority;
      pthread_setschedparam(_thread, new_policy, &param);
    }

    virtual void *run(void) = 0;

    friend void *ThreadBase_run(void *obj);

  public:

    ThreadBase(void)
    : _running(false)
    , _detach(false)
    {
    }

    virtual ~ThreadBase(void)
    {
    }

    virtual void begin(void)
    {
      if (!_running)
      {
        _detach  = false;
        _running = (pthread_create(&_thread, NULL, ThreadBase_run, this) == 0);
      }
    }

    virtual void end(void)
    {
      if (_running)
      {
        _running = false;
        if (pthread_kill(_thread, SIGUSR1) == 0)
        {
          if (!_detach)
            pthread_join(_thread, NULL);
        }
      }
    }
};

#endif
