/*
  W1GPIO.h OneWire library for Linux(w1-gpio)

  Copyright (c) 2020 Sasapea's Lab. All right reserved.

  System Setting:
    [/boot/armbianEnv.txt]
    overlay = ... w1-gpio ...
    param_w1_pin=PA10
    param_w1_pin_int_pullup=1

  Referrence:
    https://www.kernel.org/doc/html/latest/w1/slaves/index.html

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/
#ifndef __W1GPIO_H
#define __W1GPIO_H

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>

#define W1GPIO_FAMILY_DEFAULT 0x00
#define W1GPIO_FAMILY_BQ27000 0x01
#define W1GPIO_FAMILY_SMEM_01 0x01
#define W1GPIO_FAMILY_SMEM_81 0x81
#define W1GPIO_THERM_DS18S20  0x10
#define W1GPIO_FAMILY_DS28E04 0x1C
#define W1GPIO_COUNTER_DS2423 0x1D
#define W1GPIO_THERM_DS1822   0x22
#define W1GPIO_EEPROM_DS2433  0x23
#define W1GPIO_THERM_DS18B20  0x28
#define W1GPIO_FAMILY_DS2408  0x29
#define W1GPIO_EEPROM_DS2431  0x2D
#define W1GPIO_FAMILY_DS2760  0x30
#define W1GPIO_FAMILY_DS2780  0x32
#define W1GPIO_FAMILY_DS2413  0x3A
#define W1GPIO_FAMILY_DS2406  0x12
#define W1GPIO_THERM_DS1825   0x3B
#define W1GPIO_FAMILY_DS2781  0x3D
#define W1GPIO_THERM_DS28EA00 0x42

#define W1GPIO_LIST_PATH  "/sys/bus/w1/devices"
#define W1GPIO_FILE_PATH  W1GPIO_LIST_PATH "/%s/%s"
#define W1GPIO_SLAVE_FILE "w1_slave"
#define W1GPIO_RW_FILE    "rw"

class W1GPIO
{
  private:

    char _names[32][16];
    size_t _count;
    int _family;

  public:

    W1GPIO(void)
    : _count(0)
    , _family(W1GPIO_FAMILY_DEFAULT)
    {
    }

    virtual ~W1GPIO(void)
    {
    }

    int search(int family = W1GPIO_FAMILY_DEFAULT)
    {
      struct dirent *dirent;
      DIR *dir;

      _count = 0;
      if ((dir = opendir(W1GPIO_LIST_PATH)) != NULL)
      {
        for (dirent = readdir(dir); dirent; dirent = readdir(dir))
        {
          if ((dirent->d_name[2] == '-') && (strlen(dirent->d_name) == 15))
          {
            int id = strtol(dirent->d_name, NULL, 16);
            if (id != W1GPIO_FAMILY_DEFAULT)
            {
              strcpy(_names[_count], dirent->d_name);
              if (++_count >= sizeof(_names) / sizeof(_names[0]))
                break;
            }
          }
        }
        closedir(dir);
      }
      return filter(family);
    }

    int filter(int family)
    {
      int rv = 0;
      _family = family;
      for (size_t i = 0; i < _count; ++i)
      {
        if ((family == W1GPIO_FAMILY_DEFAULT) || (family == strtol(_names[i], NULL, 16)))
          ++rv;
      }
      return rv;
    }

    const char *slave(int num)
    {
      for (size_t i = 0; i < _count; ++i)
      {
        if ((_family == W1GPIO_FAMILY_DEFAULT) || (_family == strtol(_names[i], NULL, 16)))
        {
          if (num-- == 0)
            return _names[i];
        }
      }
      return NULL;
    }

    int therm(const char *slave, int32_t *temp)
    {
      char buf[256];
      int rv = read(slave, W1GPIO_SLAVE_FILE, buf, sizeof(buf) - 1);
      if (rv > 0)
      {
        buf[rv] = 0;
        if (strstr(buf, "YES"))
        {
          char *p = strstr(buf, "t=");
          if (p)
          {
            *temp = atoi(p + 2);
            return 0;
          }
        }
      }
      return -1;
    }

    int read(const char *slave, const char *file, char *buf, int len)
    {
      char path[256];
      FILE *fp;
      int rv = -1;
      snprintf(path, sizeof(path), W1GPIO_FILE_PATH, slave, file);
      if ((fp = fopen(path, "r")) != NULL)
      {
        rv = fread(buf, 1, len, fp);
        fclose(fp);
      }
      return rv;
    }

    int write(const char *slave, const char *file, char *buf, int len)
    {
      char path[256];
      FILE *fp;
      int rv = -1;
      snprintf(path, sizeof(path), W1GPIO_FILE_PATH, slave, file);
      if ((fp = fopen(path, "r+")) != NULL)
      {
        rv = fwrite(buf, 1, len, fp);
        fclose(fp);
      }
      return rv;
    }
};

extern W1GPIO W1;

#endif

