/*
  Wire.h I2C library for Linux

  Copyright (c) 2020 Sasapea's Lab. All right reserved.

  System Setting:

    [Oraneg Pi Zeo][/boot/armbianEnv.txt]
    overlays=... i2c0 i2c1 ...

    /dev/i2c-0
    /dev/i2c-1

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/
#ifndef __WIRE_H
#define __WIRE_H

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdint.h>
#include <stdbool.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/i2c-dev.h>
#include <linux/i2c.h>
#include "Stream.h"
#include "Allwinner.h"

#define WIRE_DEVICE "/dev/i2c-%u"

#define WIRE_10BIT   0x8000
#define WIRE_ADDR(a) ((a) & 0x03FF)
#define WIRE_FLAG(a) ((a) & 0xFF80 ? I2C_M_TEN : 0)

#define WIRE_STATUS_OK        0
#define WIRE_STATUS_DATA_SIZE 1
#define WIRE_STATUS_ADDR_NACK 2
#define WIRE_STATUS_DATA_NACK 3
#define WIRE_STATUS_ERROR     4

#define WIRE_MAX_MSGS 8
#define WIRE_BUF_SIZE 32
#ifdef ARDUINO
  #define WIRE_MAX_BUF WIRE_BUF_SIZE
#else
  #define WIRE_MAX_BUF (sizeof(_buffer) - _offset)
#endif

class TwoWire : public Stream
{
  private:

    uint16_t _offset;
    uint16_t _index;
    uint16_t _count;
    uint8_t  _buffer[WIRE_MAX_MSGS * WIRE_BUF_SIZE];
    uint16_t _flags;
    struct i2c_msg _msgs[WIRE_MAX_MSGS];
    struct i2c_rdwr_ioctl_data _rdwr;
    uint32_t _frequency;
    uint8_t  _num;
    int _fd;

    uint8_t transfer(void)
    {
      uint8_t rv = WIRE_STATUS_ERROR;
      _offset = 0;
      _index  = 0;
      _count  = 0;
      if (_fd != -1)
      {
        int result = ioctl(_fd, I2C_RDWR, &_rdwr);
        if (result == (int)_rdwr.nmsgs)
        {
          // gathering receive bytes
          for (int i = 0; i < result; ++i)
          {
            if (_msgs[i].flags & I2C_M_RD)
            {
              memcpy(_buffer + _count, _msgs[i].buf, _msgs[i].len);
              _count += _msgs[i].len;
            }
          }
          rv = WIRE_STATUS_OK;
        }
        else if (result == -1)
          rv = WIRE_STATUS_ADDR_NACK;
        else
          rv = WIRE_STATUS_DATA_NACK;
      }
      _rdwr.nmsgs = 0;
      return rv;
    }

  public:

    TwoWire(uint8_t num)
    : _offset(0)
    , _index(0)
    , _count(0)
    , _flags(0)
    , _frequency(100000)
    , _num(num)
    , _fd(-1)
    {
      setTimeout(0);
      _rdwr.msgs  = _msgs;
      _rdwr.nmsgs = 0;
    }

    virtual ~TwoWire()
    {
    }

    void begin(bool tenbit = false)
    {
      if (_fd == -1)
      {
        char name[32];
        snprintf(name, sizeof(name), WIRE_DEVICE, _num);
        if ((_fd = open(name, O_RDWR)) == -1)
          fprintf(stderr, "[ERROR] open(%s) = %s\n", name, strerror(errno));
        _flags = tenbit ? I2C_M_TEN : 0;
        setClock(_frequency);
      }
    }

    void end(void)
    {
      if (_fd != -1)
      {
        close(_fd);
        _fd = -1;
      }
    }

    void setClock(uint32_t clock)
    {
      _frequency = clock;
      TWI0[_num].setBusClock(clock);
    }

    void beginTransmission(uint16_t address)
    {
      _count = 0;
      _index = 0;
      if (!getWriteError())
      {
        if (_rdwr.nmsgs >= sizeof(_msgs) / sizeof(_msgs[0]))
          setWriteError();
        else
        {
          _msgs[_rdwr.nmsgs].addr  = WIRE_ADDR(address);
          _msgs[_rdwr.nmsgs].flags = WIRE_FLAG(address) | _flags;
          _msgs[_rdwr.nmsgs].len   = 0;
          _msgs[_rdwr.nmsgs].buf   = _buffer + _offset;
        }
      }
    }

    uint8_t endTransmission(uint8_t sendStop = true)
    {
      uint8_t rv = WIRE_STATUS_OK;
      if (getWriteError())
      {
        if (sendStop)
        {
          clearWriteError();
          _offset = 0;
          _rdwr.nmsgs = 0;
          rv = WIRE_STATUS_DATA_SIZE;
        }
      }
      else
      {
        _msgs[_rdwr.nmsgs++].len = _index;
        _offset += _index;
        if (sendStop)
          rv = transfer();
      }
      return rv;
    }

    uint16_t requestFrom(uint16_t address, uint16_t quantity, uint8_t sendStop = true)
    {
      _index = 0;
      _count = 0;
      if (getWriteError() || (quantity > WIRE_MAX_BUF) || (_rdwr.nmsgs >= sizeof(_msgs) / sizeof(_msgs[0])))
      {
        if (!sendStop)
          setWriteError();
        else
        {
          clearWriteError();
          _offset = 0;
          _rdwr.nmsgs = 0;
        }
      }
      else
      {
        _msgs[_rdwr.nmsgs].addr  = WIRE_ADDR(address);
        _msgs[_rdwr.nmsgs].flags = WIRE_FLAG(address) | _flags | I2C_M_RD;
        _msgs[_rdwr.nmsgs].len   = quantity;
        _msgs[_rdwr.nmsgs++].buf = _buffer + _offset;
        _offset += quantity;
        if (sendStop)
          transfer();
      }
      return _count;
    }

    void flush(void)
    {
    }

    int available(void)
    {
      return _count - _index;
    }

    int peek(void)
    {
      return available() > 0 ? _buffer[_offset + _index] : -1;
    }

    int read(void)
    {
      return available() > 0 ? _buffer[_offset + _index++] : -1;
    }

    size_t write(uint8_t b)
    {
      if (!getWriteError())
      {
        if (_index < WIRE_MAX_BUF)
        {
          _buffer[_offset + _index++] = b;
          return 1;
        }
        setWriteError();
      }
      return 0;
    }
};

extern TwoWire Wire;
extern TwoWire Wire1;
extern TwoWire Wire2;
extern TwoWire Wire3;
extern TwoWire Wire4;

#endif

