/*
  sysfs_gpio.h - SYSFS GPIO Class

  Copyright (c) 2020 Sasapea's Lab. All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
*/
#ifndef __SYSFS_GPIO_H
#define __SYSFS_GPIO_H

#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/poll.h>

typedef enum
{
  SYSFS_GPIO_ACTIVE_LOW,
  SYSFS_GPIO_ACTIVE_HIGH,
} SYSFS_GPIO_POLARITY;

typedef enum
{
  SYSFS_GPIO_DIRECTION_IN,
  SYSFS_GPIO_DIRECTION_OUT,
  SYSFS_GPIO_DIRECTION_LOW,
  SYSFS_GPIO_DIRECTION_HIGH,
} SYSFS_GPIO_DIRECTION;

typedef enum
{
  SYSFS_GPIO_EDGE_NONE,
  SYSFS_GPIO_EDGE_FALLING,
  SYSFS_GPIO_EDGE_RISING,
  SYSFS_GPIO_EDGE_BOTH,
} SYSFS_GPIO_EDGE;

typedef void (*GPIO_USER_FUNC_T)(void);

#define SYSFS_GPIO_PATH           "/sys/class/gpio"
#define SYSFS_GPIO_EXPORT_PATH    SYSFS_GPIO_PATH "/export"
#define SYSFS_GPIO_UNEXPORT_PATH  SYSFS_GPIO_PATH "/unexport"
#define SYSFS_GPIO_PORT_PATH      SYSFS_GPIO_PATH "/gpio%u"
#define SYSFS_GPIO_POLARITY_PATH  SYSFS_GPIO_PORT_PATH "/active_low"
#define SYSFS_GPIO_DIRECTION_PATH SYSFS_GPIO_PORT_PATH "/direction"
#define SYSFS_GPIO_EDGE_PATH      SYSFS_GPIO_PORT_PATH "/edge"
#define SYSFS_GPIO_VALUE_PATH     SYSFS_GPIO_PORT_PATH "/value"
#define SYSFS_GPIO_PATH_MAX       64

class sysfs_gpio
{
  private:

    static int write(const char *name, uint16_t pin, const char *buf, size_t len)
    {
      int fd, rv = -1;
      char path[SYSFS_GPIO_PATH_MAX];
      snprintf(path, sizeof(path), name, pin);
      if ((fd = open(path, O_WRONLY)) != -1)
      {
        rv = ::write(fd, buf, len);
        close(fd);
      }
      return rv;
    }

    static int read(const char *name, uint16_t pin, char *buf, size_t len)
    {
      int fd, rv = -1;
      char path[SYSFS_GPIO_PATH_MAX];
      snprintf(path, sizeof(path), name, pin);
      if ((fd = open(path, O_RDONLY)) != -1)
      {
        rv = ::read(fd, buf, len);
        close(fd);
      }
      return rv;
    }

    static int poll(const char *name, uint16_t pin, GPIO_USER_FUNC_T func, int timeout)
    {
      struct pollfd fds[1];
      int rv = -1;
      char path[SYSFS_GPIO_PATH_MAX];
      snprintf(path, sizeof(path), name, pin);
      if ((fds[0].fd = open(path, O_RDONLY)) != -1)
      {
        ::read(fds[0].fd, path, 1);
        fds[0].events = POLLPRI;
        while (1)
        {
          rv = ::poll(fds, sizeof(fds) / sizeof(fds[0]), timeout);
          if ((rv <= 0) || !func)
            break;
          lseek(fds[0].fd, 0, SEEK_SET);
          ::read(fds[0].fd, path, 1);
          func();
        }
        close(fds[0].fd);
      }
      return rv;
    }

    sysfs_gpio(void)
    {
    }

    virtual ~sysfs_gpio(void)
    {
    }

  public:

    static int begin(uint16_t pin)
    {
      int rv = -1;
      char buf[16];
      int len = snprintf(buf, sizeof(buf), "%u", pin);
      if (write(SYSFS_GPIO_EXPORT_PATH, pin, buf, len) == len)
        rv = 0;
      return rv;
    }

    static int end(uint16_t pin)
    {
      int rv = -1;
      char buf[16];
      int len = snprintf(buf, sizeof(buf), "%u", pin);
      if (write(SYSFS_GPIO_UNEXPORT_PATH, pin, buf, len) == len)
        rv = 0;
      return rv;
    }

    static int polarity(uint16_t pin, SYSFS_GPIO_POLARITY mode)
    {
      static const char *MODE[] = {"0", "1"};
      int rv = -1;
      if (mode < sizeof(MODE) / sizeof(MODE[0]))
      {
        const char *buf = MODE[mode];
        int len = strlen(buf);
        if (write(SYSFS_GPIO_POLARITY_PATH, pin, buf, len) == len)
          rv = 0;
      }
      return rv;
    }

    static int direction(uint16_t pin, SYSFS_GPIO_DIRECTION mode)
    {
      static const char *MODE[] = {"in", "out", "low", "high"};
      int rv = -1;
      if (mode < sizeof(MODE) / sizeof(MODE[0]))
      {
        const char *buf = MODE[mode];
        int len = strlen(buf);
        if (write(SYSFS_GPIO_DIRECTION_PATH, pin, buf, len) == len)
          rv = 0;
      }
      return rv;
    }

    static int edge(uint16_t pin, SYSFS_GPIO_EDGE mode)
    {
      static const char *MODE[] = {"none", "falling", "rising", "both"};
      int rv = -1;
      if (mode < sizeof(MODE) / sizeof(MODE[0]))
      {
        const char *buf = MODE[mode];
        int len = strlen(buf);
        if (write(SYSFS_GPIO_EDGE_PATH, pin, buf, len) == len)
          rv = 0;
      }
      return rv;
    }

    static void write(uint16_t pin, int value)
    {
      char c = value ? '1' : '0';
      write(SYSFS_GPIO_VALUE_PATH, pin, &c, sizeof(c));
    }

    static int read(uint16_t pin)
    {
      char c = '0';
      read(SYSFS_GPIO_VALUE_PATH, pin, &c, sizeof(c));
      return c & 1;
    }

    static int poll(uint16_t pin, GPIO_USER_FUNC_T func, int timeout)
    {
      return poll(SYSFS_GPIO_VALUE_PATH, pin, func, timeout);
    }
};

#endif
